#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) 2019 Chintalagiri Shashank
#
# This file is part of tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Reusable Schema Elements (:mod:`tendril.schema.helpers`)
========================================================
"""

from six import iteritems
from inspect import isclass

from collections import MutableMapping
from collections import MutableSequence

from tendril.validation.base import ValidatableBase
from tendril.validation.base import ValidationError
from tendril.validation.files import ExtantFile

try:
    from tendril.utils.types import ParseException
    _exc = (ParseException, ValidationError)
except ImportError:
    _exc = ValidationError


class MultilineString(list):
    def __init__(self, value):
        super(MultilineString, self).__init__(value)

    def __repr__(self):
        return '\n'.join(self)


class SchemaObjectCollection(ValidatableBase):
    _objtype = None
    _pass_args = []
    _validator = None
    _allow_empty = True

    def __init__(self, content, *args, **kwargs):
        for arg in self._pass_args:
            setattr(self, arg, kwargs.pop(arg))
        super(SchemaObjectCollection, self).__init__(*args, **kwargs)
        self._source_content = content
        self._content = self._empty_container

    @property
    def _empty_container(self):
        raise NotImplementedError

    @property
    def content(self):
        return self._content

    def _parse_item_with(self, item, objtype):
        if isclass(objtype) and \
                issubclass(objtype, ValidatableBase):
            if self._pass_args:
                args = {k: getattr(self, k) for k in self._pass_args}
                value = objtype(item, vctx=self._validation_context, **args)
            else:
                value = objtype(item, vctx=self._validation_context)
            value.validate()
            self._validation_errors.add(value.validation_errors)
        elif objtype:
            value = objtype(item)
        else:
            value = item
        return value

    def _parse_item(self, item):
        if isinstance(self._objtype, list):
            default_parser = None
            for sig, parser in self._objtype:
                if sig == 'default':
                    default_parser = parser
                    continue
                if isinstance(item, sig):
                    return self._parse_item_with(item, parser)
            return self._parse_item_with(item, default_parser)
        else:
            return self._parse_item_with(item, self._objtype)

    def _validate_item(self, item):
        if not item:
            return False
        if self._validator:
            try:
                self._validator(item)
                return True
            except _exc as e:
                self._validation_errors.add(e)
                return False
        return True

    def _validate(self):
        pass

    def __iter__(self):
        return iter(self._content)

    def __getitem__(self, item):
        return self._content[item]

    def __setitem__(self, key, value):
        self._content[key] = value

    def __delitem__(self, key):
        self._content.__delitem__(key)

    def __len__(self):
        return len(self._content)


class SchemaObjectList(SchemaObjectCollection, MutableSequence):
    def __init__(self, *args, **kwargs):
        super(SchemaObjectList, self).__init__(*args, **kwargs)
        if not self._source_content and self._allow_empty:
            return
        for item in self._source_content:
            if not self._validate_item(item):
                continue
            self._content.append(self._parse_item(item))

    @property
    def _empty_container(self):
        return []

    def get(self, item):
        if not self._objtype or not hasattr(self._objtype, 'handle'):
            raise NotImplementedError("handle not specified for {0}"
                                      "".format(self.__class__.__name__))

        for candidate in self.content:
            if getattr(candidate, self._objtype.handle) == item:
                return candidate

        raise ValueError("{0} with handle {1} not found."
                         "".format(self._objtype.__name__, item))

    @property
    def handles(self):
        if not self._objtype or not hasattr(self._objtype, 'handle'):
            raise NotImplementedError("handle not specified for {0}"
                                      "".format(self.__class__.__name__))
        return [getattr(x, self._objtype.handle) for x in self.content]

    def insert(self, index, item):
        return self._content.insert(index, item)


class FileList(SchemaObjectList):
    _objtype = ExtantFile
    _pass_args = ['basedir']


class SchemaObjectMapping(SchemaObjectCollection, MutableMapping):
    def __init__(self, *args, **kwargs):
        super(SchemaObjectMapping, self).__init__(*args, **kwargs)
        if not self._source_content and self._allow_empty:
            return
        for k, v in iteritems(self._source_content):
            if not self._validate_item(v):
                continue
            self._content[k] = self._parse_item(v)

    @property
    def _empty_container(self):
        return {}

    def keys(self):
        return self.content.keys()


class SchemaSelectableObjectMapping(SchemaObjectMapping):
    def __init__(self, content, *args, **kwargs):
        default = content.pop('default')
        super(SchemaSelectableObjectMapping, self).__init__(content, *args, **kwargs)
        self.default = self.content[default]

    def __getitem__(self, item):
        if not item:
            return self.default
        return super(SchemaSelectableObjectMapping, self).__getitem__(item)

    def __repr__(self):
        return "<{0} {1}>".format(self.__class__.__name__ ,
                                  ','.join(self.content.keys()))


def load(manager):
    pass



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) 2016-2019 Chintalagiri Shashank
#
# This file is part of tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Validation elements for Schema Declarations
===========================================
"""


from tendril.validation.base import ValidationPolicy
from tendril.validation.base import ValidationError


class SchemaPolicy(ValidationPolicy):
    def __init__(self, context, name, vmax, vmin):
        super(SchemaPolicy, self).__init__(context)
        self.name = name
        self.vmax = vmax
        self.vmin = vmin

    def validate(self, name, version):
        if name == self.name and self.vmin <= version <= self.vmax:
            return True
        else:
            return False

    def render(self):
        return "Supports {0}<={1}<={2}".format(self.vmin, self.name, self.vmax)


class SchemaNotSupportedError(ValidationError):
    msg = "The file specifies a schema which is not supported."

    def __init__(self, policy, value):
        super(SchemaNotSupportedError, self).__init__(policy)
        self._value = value

    def __repr__(self):
        return "<SchemaNotSupportedError {0} {1}>" \
               "".format(self.policy.context, self._value)

    def render(self):
        return {
            'is_error': self.policy.is_error,
            'group': self.msg,
            'headline': "Got {0} in {1}".format(
                self._value, self.policy.context),
            'detail': self.policy.render(),
        }



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) 2016-2019 Chintalagiri Shashank
#
# This file is part of tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Validation elements for Schema Controlled Files
===============================================
"""


from inspect import isclass
from tendril.validation.base import ValidatableBase
from tendril.validation.base import ValidationError
from tendril.validation.base import ValidationPolicy


class ContextualConfigError(ValidationError):
    msg = "Incorrect Configuration"

    def __init__(self, policy):
        super(ContextualConfigError, self).__init__(policy)

    def _format_path(self):
        if isinstance(self._policy.path, tuple):
            return '/'.join(self._policy.path)
        else:
            return self._policy.path

    def render(self):
        return {
            'is_error': self.policy.is_error,
            'group': self.msg,
            'headline': self._policy.context.render(),
            'detail': "Configuration seems to be incorrect.",
        }


class ConfigKeyError(ContextualConfigError):
    msg = "Configuration Key Missing"

    def __init__(self, policy):
        super(ConfigKeyError, self).__init__(policy)

    def __repr__(self):
        return "<ConfigKeyError {0} {1}>" \
               "".format(self._policy.context, self._format_path())

    def render(self):
        if self._policy.options:
            option_str = "Valid options are {0}" \
                         "".format(', '.join(self._policy.options))
        else:
            option_str = ''
        return {
            'is_error': self.policy.is_error,
            'group': self.msg,
            'headline': "{0} missing in {1}"
                        "".format(self._format_path(),
                                  self._policy.context.render()),
            'detail': "This required configuration option could not be "
                      "found in the configs file. " + option_str,
        }


class ConfigValueInvalidError(ContextualConfigError):
    msg = "Configuration Value Unrecognized"

    def __init__(self, policy, value):
        super(ConfigValueInvalidError, self).__init__(policy)
        self._value = value

    def __repr__(self):
        return "<ConfigValueInvalidError {0} {1}>" \
               "".format(self._policy.context, self._format_path())

    def render(self):
        if self._policy.options:
            option_str = "Valid options are {0}" \
                         "".format(', '.join(self._policy.options))
        else:
            option_str = ''
        return {
            'is_error': self.policy.is_error,
            'group': self.msg,
            'headline': "'{0}' Invalid for {1} in {2}"
                        "".format(self._value, self._format_path(),
                                  self._policy.context.render()),
            'detail': "The value provided for this configuration option is "
                      "unrecognized or not allowed in this context. " +
                      option_str,
        }


class ConfigOptionPolicy(ValidationPolicy):
    def __init__(self, context, path, parser=None, parser_args=None,
                 required=True, options=None, default=None, is_error=True):
        super(ConfigOptionPolicy, self).__init__(context, is_error)
        self.path = path
        self.parser = parser
        self._parser_args = parser_args
        self.options = options
        self.default = default
        self.required = required

    @property
    def parser_args(self):
        return self._parser_args or {}

    def get(self, data):
        if self.path is None:
            return data
        try:
            return get_dict_val(data, self)
        except ConfigKeyError as error:
            if not self.required:
                if self.default is None or not self.parser:
                    return self.default
                if isclass(self.parser) and isinstance(self.default, self.parser):
                    return self.default
                vctx = self.context.child(self.parser.__name__)
                return _parse(self.parser, self.default, vctx=vctx, **self.parser_args)
            else:
                raise error


def _parse(parser, value, vctx=None, **parser_args):
    if isclass(parser) and \
            issubclass(parser, ValidatableBase):
        return parser(value, vctx=vctx, **parser_args)
    else:
        return parser(value)


def get_dict_val(d, policy=None):
    try:
        assert isinstance(d, dict)
    except AssertionError:
        print("Expected to get a dictionary here. This probably means the YAML "
              "file is empty or unrecognizably mangled. Got {0} instead.".format(d))
        print(policy.context)
        raise
    if isinstance(policy.path, tuple):
        try:
            for key in policy.path:
                if key not in d.keys():
                    raise KeyError
                d = d.get(key)
        except (KeyError, AttributeError):
            raise ConfigKeyError(policy=policy)
        rval = d
    else:
        try:
            if policy.path not in d.keys():
                raise KeyError
            rval = d.get(policy.path)
        except KeyError:
            raise ConfigKeyError(policy=policy)

    if policy.parser:
        try:
            if isinstance(policy.parser, tuple):
                for parser in policy.parser:
                    try:
                        vctx = policy.context.child(parser.__name__)
                        rval = _parse(parser, rval, vctx, **policy.parser_args)
                        break
                    except:
                        continue
                else:
                    raise Exception
            else:
                vctx = policy.context.child(policy.parser.__name__)
                rval = _parse(policy.parser, rval, vctx, **policy.parser_args)
        except Exception as e:
            raise ConfigValueInvalidError(policy=policy, value=rval)

    if policy.options is None or rval in policy.options:
        return rval
    else:
        raise ConfigValueInvalidError(policy=policy, value=rval)



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) 2019 Chintalagiri Shashank
#
# This file is part of tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Base Schemas (:mod:`tendril.schema.base`)
=========================================
"""

import os
import warnings
from six import iteritems
from decimal import Decimal
from jinja2 import Template
from tendril.utils.files import yml as yaml

from tendril.validation.base import ValidatableBase
from tendril.validation.base import ValidationContext
from tendril.validation.schema import SchemaPolicy
from tendril.validation.schema import SchemaNotSupportedError
from tendril.validation.configs import ConfigOptionPolicy
from tendril.validation.configs import ContextualConfigError

from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)


class SchemaProcessorBase(ValidatableBase):
    def __init__(self, *args, **kwargs):
        super(SchemaProcessorBase, self).__init__(*args, **kwargs)
        self._policies = {}
        self._load_schema_policies()

    @property
    def _raw(self):
        return self._raw_content

    def _p(self, *args, **kwargs):
        return ConfigOptionPolicy(self._validation_context, *args, **kwargs)

    def elements(self):
        return {}

    def schema_policies(self):
        policies = self.elements()
        return policies

    def _load_schema_policies(self):
        self._policies.update(self.schema_policies())

    def _process_element(self, key, policy):
        if isinstance(policy, ConfigOptionPolicy):
            try:
                value = policy.get(self._raw)
                if isinstance(value, ValidatableBase):
                    value.validate()
                    self._validation_errors.add(value.validation_errors)
                setattr(self, key, value)
            except ContextualConfigError as e:
                self._validation_errors.add(e)

    def _process(self):
        for key, policy in iteritems(self._policies):
            self._process_element(key, policy)

    def __getattr__(self, item):
        if item not in self._policies.keys():
            raise AttributeError("%r has no attribute %r" % (type(self), item))
        policy = self._policies[item]
        return policy.get(self._raw)

    def _validate(self):
        self._validated = True


class NakedSchemaObject(SchemaProcessorBase):
    def __init__(self, content, *args, **kwargs):
        super(NakedSchemaObject, self).__init__(*args, **kwargs)
        self._raw_content = content
        self._process()
        if self.validation_errors.terrors:
            warnings.warn("{0} of class {1} has {2} Validation Errors"
                          "".format(self.ident, self.__class__.__name__,
                                    self.validation_errors.terrors),
                          UserWarning)


class SchemaControlledObject(NakedSchemaObject):
    legacy_schema_name = None
    supports_schema_name = None
    supports_schema_version_max = None
    supports_schema_version_min = None

    def __init__(self, *args, strict_schema=False, **kwargs):
        self._strict_schema = strict_schema
        super(SchemaControlledObject, self).__init__(*args, **kwargs)

    def _stub_content(self):
        return {
            'schema_name': self.supports_schema_name,
            'schema_version': self.supports_schema_version_max,
        }

    def elements(self):
        e = super(SchemaControlledObject, self).elements()
        e.update({
            'schema_name':    self._p(('schema', 'name'),),
            'schema_version': self._p(('schema', 'version'), parser=Decimal),
        })
        return e

    def schema_policies(self):
        policies = super(SchemaControlledObject, self).schema_policies()
        policies.update({
                'schema_policy': SchemaPolicy(
                    self._validation_context,
                    self.supports_schema_name,
                    self.supports_schema_version_max,
                    self.supports_schema_version_min
                )
        })
        return policies

    def _verify_schema_decl(self):
        policy = self._policies['schema_policy']
        if self.supports_schema_name == '*':
            return
        if self.schema_name == self.legacy_schema_name:
            self.schema_name = self.supports_schema_name
        logger.debug("Validating Schema Policy : {0} {1}"
                     "".format(self.schema_name, self.schema_version))
        if not policy.validate(self.schema_name, self.schema_version):
            raise SchemaNotSupportedError(
                policy,
                '{0} v{1}'.format(self.schema_name, self.schema_version)
            )

    def _process(self):
        super(SchemaControlledObject, self)._process()
        try:
            self._verify_schema_decl()
        except SchemaNotSupportedError as e:
            if self._strict_schema:
                raise
            self._validation_errors.add(e)


class SchemaControlledYamlFile(SchemaControlledObject):
    supports_schema_name = '*'
    FileNotFoundExceptionType = None
    template = None

    def __init__(self, path, *args, **kwargs):
        self._path = path
        vctx = ValidationContext(
            self._path,
            locality=self.supports_schema_name or self.__class__.__name__
        )
        raw_content = self._get_yaml_file()
        super(SchemaControlledYamlFile, self).__init__(
            raw_content, *args, vctx=vctx, **kwargs
        )

    @property
    def path(self):
        return self._path

    def _generate_stub(self):
        template = Template(open(self.template).read())
        with open(self._path, 'w') as f:
            f.write(template.render(stage=self._stub_content()))

    def _get_yaml_file(self):
        if self.template and not os.path.exists(self._path):
            self._generate_stub()
        if self.FileNotFoundExceptionType and not os.path.exists(self._path):
            raise self.FileNotFoundExceptionType(self._path)
        return yaml.load(self._path)


def load(manager):
    logger.debug("Loading {0}".format(__name__))
    manager.load_schema('SchemaControlledYamlFile', SchemaControlledYamlFile,
                        doc="Base class for schema controlled file processors.")



#!/usr/bin/env python
# encoding: utf-8

# Copyright (C) 2019 Chintalagiri Shashank
#
# This file is part of tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Tendril Schema Manager (:mod:`tendril.schema.manager`)
======================================================
"""


import importlib

from tendril.validation.base import ValidationContext
from tendril.validation.configs import ConfigOptionPolicy
from tendril.validation.schema import SchemaNotSupportedError
from tendril.schema.base import SchemaControlledYamlFile

from tendril.utils.versions import get_namespace_package_names
from tendril.utils import log
logger = log.get_logger(__name__, log.DEBUG)


class SchemaManager(object):
    def __init__(self, prefix):
        self._prefix = prefix
        self._schemas = {}
        self._file_schemas = {}
        self._docs = []
        self._load_schemas()
        self._validation_context = ValidationContext(self.__module__)

    def _load_schemas(self):
        logger.debug("Loading schema modules from {0}".format(self._prefix))
        modules = list(get_namespace_package_names(self._prefix))
        for m_name in modules:
            if m_name == __name__:
                continue
            m = importlib.import_module(m_name)
            m.load(self)
        logger.debug("Done loading schema modules from {0}".format(self._prefix))

    def load_schema(self, name, processor, doc):
        logger.debug("Installing schema definition {0}".format(name))
        self._schemas[name] = processor
        if issubclass(processor, SchemaControlledYamlFile):
            self._file_schemas[name] = processor
        self._docs.append((name, doc))

    def __getattr__(self, item):
        if item == '__file__':
            return None
        if item == '__path__':
            return None
        if item == '__len__':
            return len(self._schemas.keys())
        if item == '__all__':
            return list(self._schemas.keys()) + \
                   ['load_schema', 'load', 'doc_render']
        return self._schemas[item]

    def load(self, targetpath):
        baseparser = getattr(self, 'SchemaControlledYamlFile')
        target = baseparser(targetpath)
        target_schema = target.schema_name
        if target_schema not in self._file_schemas.keys():
            # TODO Replace with a generic OptionPolicy?
            policy = ConfigOptionPolicy(self._validation_context,
                                        'schema.name',
                                        self._file_schemas.keys())
            raise SchemaNotSupportedError(policy, target_schema)
        return getattr(self, target_schema)(targetpath)

    def doc_render(self):
        return self._docs

    def __repr__(self):
        return "<SchemaManager>"
